"""Default logger of pytorch lightning."""
from typing import Any

from cthunder.loggers.logger_abstract import LoggerAbstract


class DefaultLogger(LoggerAbstract):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init_logger(self, *args, **kwargs):
        """Nothing to do for initialisation."""
        pass

    def log_loss(self, split: str, loss: Any, *args, **kwargs):
        """Log the loss."""
        # Get the model from kwargs
        pl_model = kwargs.get("pl_model", None)
        if pl_model is not None:
            pl_model.log(f"{split}_loss", loss, prog_bar=True)

    def log_metric(self, split: str, score: Any, metric_name: str, *args, **kwargs):
        """Log the metric."""
        # Get the model from kwargs
        pl_model = kwargs.get("pl_model", None)
        if pl_model is not None:
            pl_model.log(f"{split}_{metric_name}_step", score)

    def log_metric_end_epoch(
        self, split: str, score: Any, metric_name: str, *args, **kwargs
    ):
        """Log at the end of epoch."""
        # Get the model from kwargs
        pl_model = kwargs.get("pl_model", None)
        if pl_model is not None:
            pl_model.log(f"{split}_{metric_name}_total", score)



"""Weight&Biases logger."""
import os
from typing import Any, Optional, Dict

from dotenv import load_dotenv

import wandb
from omegaconf import OmegaConf

from cthunder.loggers.logger_abstract import LoggerAbstract


class WandbLogger(LoggerAbstract):
    """Wandb custom logger."""

    def __init__(self, name: str = "rna3d", is_local: bool = False, *args, **kwargs):
        """
        Should have in `.env'. file the given key : WB_LOCAL_KEY
        """
        self.name = name
        self.is_local = is_local
        super().__init__(*args, **kwargs)

    def init_logger(self, config: Optional[Dict] = {},*args, **kwargs):
        """Init the logger"""
        # Load the variables in .env file
        load_dotenv()
        wb_token = os.environ.get("WANDB_API_KEY")
        os.environ["WANDB_API_KEY"] = wb_token
        config = OmegaConf.to_container(config, resolve=True)
        mode = "offline" if self.is_local else "online"
        wandb.init(project="rnattention", name=self.name, config=config, mode=mode, dir="logs")

    def log_loss(self, split: str, loss: Any, *args, **kwargs):
        """Log the loss."""
        wandb.log({f"{split}_loss": loss})

    def log_metric(self, split: str, score: Any, metric_name: str, *args, **kwargs):
        """Log the metric."""
        wandb.log({f"{split}_{metric_name}": score})

    def log_metric_end_epoch(
        self, split: str, score: Any, metric_name: str, *args, **kwargs
    ):
        """Log at the end of epoch."""
        wandb.log({f"{split}_{metric_name}_epoch": score})

    def log_value(self, name: str, value: Any, *args, **kwargs):
        wandb.log({name: value})



"""
Class that implements the abstract interface for the postprocessing.
"""
from abc import abstractmethod

from cthunder.data.sample import Sample


class PostprocessAbstract:
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def transform(self, obj: Sample):
        raise NotImplementedError



"""Class to define the different logging functions."""
from abc import abstractmethod
from typing import Any


class LoggerAbstract:
    def __init__(self, *args, **kwargs):
        self.init_logger(*args, **kwargs)

    @abstractmethod
    def init_logger(self, *args, **kwargs):
        """Init the logger"""
        raise NotImplementedError

    @abstractmethod
    def log_loss(self, split: str, loss: Any, *args, **kwargs):
        """Log the loss."""
        raise NotImplementedError

    @abstractmethod
    def log_metric(self, split: str, score: Any, metric_name: str, *args, **kwargs):
        """Log the metric."""
        raise NotImplementedError

    @abstractmethod
    def log_metric_end_epoch(
        self, split: str, score: Any, metric_name: str, *args, **kwargs
    ):
        """Log at the end of epoch."""
        raise NotImplementedError



"""Abstract PytorchLightningModule class."""
from typing import Any, Dict, List, Optional

import torch
from loguru import logger
from lightning.pytorch import LightningModule
from omegaconf import OmegaConf
from torch import nn
from torchmetrics import Accuracy, F1Score, Precision, Recall

from cthunder.features.postprocess_abstract import PostprocessAbstract
from cthunder.loggers.default_logger import DefaultLogger
from cthunder.loggers.logger_abstract import LoggerAbstract
from cthunder.utils.utils import instantiate_class_from_init


class AbstractPlModule(LightningModule):
    """
    Class that defines all the needed functions to be used for deep learning approches.
    """

    def __init__(
        self,
        model: nn.Module,
        custom_loggers: Optional[List[LoggerAbstract]] = None,
        loss_args: Any = None,
        postprocesses: Optional[List[PostprocessAbstract]] = None,
        checkpoint_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        Initialise the model, hyperparameters, loss.
        :param model: the pytorch model to use
        :param custom_loggers: a list of custom loggers.
        :param loss_args: the path to the loss to use for the training.
        :param checkpoint_path: the path to the checkpoint to load.
        """
        super().__init__()
        self.model = model
        self.custom_loggers = (
            custom_loggers if custom_loggers is not None else [DefaultLogger()]
        )
        # Variables to fill
        self.metrics_dict_train: Dict[str, Any] = {}
        self.metrics_dict_val: Dict[str, Any] = {}
        self.metrics_dict_test: Dict[str, Any] = {}
        self.config_path = (
            None  # Used only to keep track of the path of the config file
        )
        self.loss = instantiate_class_from_init(loss_args)
        # Setups
        self.setup_hp(*args, **kwargs)
        self.setup_metrics(*args, **kwargs)
        self.postprocesses = postprocesses

    def forward(self, x, y):
        output = self.model(x)
        return output

    def training_step(self, batch, batch_nb):
        """Compute the logits and the loss."""
        x, y = self.interpret_batch(batch)
        logits = self.forward(x, y)
        loss = self.loss(logits, y)
        preds = self.postprocess_eval(logits)
        self.log_metrics(y, preds, loss, "train")
        return loss

    def interpret_batch(self, batch):
        """Convert batch to x and y"""
        x, y = batch
        return x, y

    def setup_hp(self, lr: float, optimizer_args: Dict, scheduler_args: Optional[Dict] = None, *args, **kwargs):
        """Set the hyperparameters.
        Args:
            :param lr: learning rate
            :param optimizer_args: dictionary with the class path and init args
        """
        self.lr = lr
        self.optimizer_args = OmegaConf.to_container(optimizer_args)  # convert to dict
        self.scheduler_args = OmegaConf.to_container(scheduler_args) if scheduler_args is not None else None

    def configure_optimizers(self):
        """Configure the optimizer"""
        if self.optimizer_args is None:
            logger.debug("OPTIMIZER NAME NOT FOUND")
        self.optimizer_args["init_args"]["params"] = self.model.parameters()
        optimizer = instantiate_class_from_init(self.optimizer_args)
        if self.scheduler_args is not None:
            scheduler_args = self.scheduler_args
            scheduler_args["init_args"]["optimizer"] = optimizer
            scheduler = instantiate_class_from_init(self.scheduler_args)
            return [optimizer], [scheduler]
        return optimizer

    def setup_metrics(self, num_classes: int = 2, *args, **kwargs):
        """Set the different metrics.

        Args:
             :param num_classes: number of labels
        """
        self.metrics_dict_train = self._get_metrics(num_classes)
        self.metrics_dict_test = self._get_metrics(num_classes)
        self.metrics_dict_val = self._get_metrics(num_classes)

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        """Do prediction for the given dataset"""
        x, y = self.interpret_batch(batch)
        logits = self.forward(x, y)
        preds = self.postprocess_eval(logits, dim=1).cpu().numpy()
        return {"y": y.cpu().numpy(), "preds": preds}

    def _get_metrics(self, num_classes: int) -> nn.ModuleDict:
        """
        Return a ModuleDict of different metrics.
        Args:
            :param num_classes: number of labels
        """
        task = "binary" if num_classes == 2 else "multiclass"
        return nn.ModuleDict(
            {
                "Accuracy": Accuracy(task=task, num_classes=num_classes),
                "F1 score": F1Score(task=task, num_classes=num_classes),
                "Recall": Recall(task=task, num_classes=num_classes),
                "Precision": Precision(task=task, num_classes=num_classes),
            }
        )

    def postprocess_eval(self, logits: torch.Tensor, dim: int = 1) -> torch.Tensor:
        """Return the argmax of the logits
        :param logits: the output of the forward.
        :param dim: the dimension where to take the argmax.
        :return the argmax for dimension 1.
        """
        preds = torch.argmax(logits, dim=dim)
        return preds

    def evaluate(self, batch, batch_nb, split: str) -> None:
        """
        :param split: which split (test or validation) for logging
        :return:
        """
        x, y = self.interpret_batch(batch)
        logits = self.forward(x, y)
        loss = self.loss(logits, y)  # type: ignore
        preds = self.postprocess_eval(logits)
        self.log_metrics(y, preds, loss, split)

    def log_metrics(self, y, preds, loss, split) -> None:
        """Log the different metrics."""
        if split in ["train", "training"]:
            metrics_dict = self.metrics_dict_train
        elif split in ["val", "valid", "validation"]:
            metrics_dict = self.metrics_dict_val
        elif split in ["test", "testing"]:
            metrics_dict = self.metrics_dict_test
        else:
            logger.debug(f"SPLIT NOT FOUND : {split}")
            return None
        self._log_metrics(y, preds, loss, split, metrics_dict)

    def _compute_scores(self, metric_dict: Dict, preds: torch.Tensor, y: torch.Tensor):
        """Compute the scores."""
        scores = {}
        for metric_name, metric in metric_dict.items():
            score = metric(preds, y)
            scores[metric_name] = score
        return scores

    def _compute_scores_epoch(self, metric_dict: Dict):
        """Compute the score at the end of an epoch. Reset also the scores after."""
        scores = {}
        for metric_name, metric in metric_dict.items():
            score = metric.compute()
            scores[metric_name] = score
            metric.reset()
        return scores

    def _log_metrics(self, y, preds, loss, split, metric_dict) -> None:
        """Log the metric with the given dictionary of metrics."""
        scores = self._compute_scores(metric_dict, preds, y)
        if self.custom_loggers is not None:
            # Loop over the different loggers
            for c_logger in self.custom_loggers:
                # Log the loss. The 'pl_model=self' is useful only for default logger
                c_logger.log_loss(split, loss, pl_model=self)
                # Log the metrics
                self._log_metrics_to_logger(c_logger, split, scores)
                if self.scheduler_args is not None:
                    lr = self.lr_schedulers().get_last_lr()[0]
                    c_logger.log_value("Learning rate", lr)

    def _log_metrics_to_logger(
        self,
        logger: LoggerAbstract,
        split: str,
        scores: Dict,
    ):
        """Log metrics to the custom logger."""
        for metric_name, score in scores.items():
            # Error if the metric isn't in the same device.
            if split == "train":
                # Log only for the training set
                logger.log_metric(split, score, metric_name, pl_model=self)

    def _log_metrics_to_logger_end_epoch(
        self, logger: LoggerAbstract, scores: Dict, split: str
    ):
        """Log metrics at the end of epoch."""
        for metric_name, score in scores.items():
            if isinstance(score, Dict):
                for m_name, c_score in score.items():
                    logger.log_metric_end_epoch(split, c_score, metric_name+"_"+m_name, pl_model=self)
            else:
                logger.log_metric_end_epoch(split, score, metric_name, pl_model=self)

    def validation_step(self, batch, batch_idx):
        self.evaluate(batch, batch_idx, "val")

    def test_step(self, batch, batch_idx):
        self.evaluate(batch, batch_idx, "test")

    def _shared_epoch_end(self, metric_dict: Dict, split: str):
        """Log at the end of the epoch."""
        if self.custom_loggers is not None:
            scores = self._compute_scores_epoch(metric_dict)
            for c_logger in self.custom_loggers:
                self._log_metrics_to_logger_end_epoch(c_logger, scores, split)

    def on_train_epoch_end(self) -> None:
        """Logging at the end of each training epoch."""
        self._shared_epoch_end(self.metrics_dict_train, "train")

    def on_validation_epoch_end(self) -> None:
        """Logging at the end of validation epoch"""
        self._shared_epoch_end(self.metrics_dict_val, "val")

    def on_test_epoch_end(self) -> None:
        """Logging at the end of test epoch"""
        self._shared_epoch_end(self.metrics_dict_test, "test")



import inspect
import os
import shutil
from typing import List, Optional

from loguru import logger
from pytorch_lightning import Trainer
from torch import nn

from cthunder.config.config_helper_abstract import ConfigHelperAbstract
from cthunder.data.datamodule_abstract import DataModuleAbstract
from cthunder.data.dataset_abstract import DatasetAbstract
from cthunder.features.preprocess_abstract import PreprocessAbstract
from cthunder.loggers.logger_abstract import LoggerAbstract
from cthunder.pl_model.abstract_pl_model import AbstractPlModule


class ConfigHelperPython(ConfigHelperAbstract):
    """Class that return the main elements from the python config file."""

    def __init__(self, config_path: Optional[str], *args, **kwargs):
        """
        Load the config.py file.
        :param path_to_config: the path to the python config file.
        """
        super().__init__(config_path, *args, **kwargs)

    def read_config(self, config_path: Optional[str]):
        """Read the content of the config file."""
        pass

    def get_preprocesses(self) -> Optional[List[PreprocessAbstract]]:
        """
        Return the preprocesses from the config file.
        :return: a list of preprocesses initialised with they parameters
        """
        classes = HPKind.PREPROCESS.get("classes", None)
        params = HPKind.PREPROCESS.get("params", None)
        if classes is not None:
            preprocesses = []
            for classe, param in zip(classes, params):
                preprocesses.append(classe(**param))
            return preprocesses
        else:
            logger.debug("CLASSES FOR PREPROCESSING IS None")
            return None

    def get_dataset(self) -> Optional[DatasetAbstract]:
        """
        Return the Dataset class from the config file.
        :return: the Dataset instantiate.
        """
        dataset_class = HPKind.DATASET.get("class", None)
        if dataset_class is not None:
            return dataset_class
        else:
            raise AttributeError("FAIL TO LOAD DATASET")

    def get_datamodule(self) -> DataModuleAbstract:
        """
        Return the DataLoader class from the config file.
        It initialises the datasets and preprocesses
        :return: the DataLoader instantiate.
        """
        # Get the preprocesses
        preprocess = self.get_preprocesses()
        # Get the dataset class
        dataset_class = self.get_dataset()
        # Get the dataset params
        dataset_params = HPKind.DATASET.get("params", {})
        # Get the dataloader params and class
        datamodule_class = HPKind.DATAMODULE.get("class", None)
        datamodule_params = HPKind.DATAMODULE.get("params", {})
        # Fill the last parameters : preprocesses and dataset class
        datamodule_params["preprocesses"] = preprocess
        datamodule_params["dataset_class"] = dataset_class
        datamodule_params = {**dataset_params, **datamodule_params}
        if datamodule_class is not None:
            dataloader = datamodule_class(
                **datamodule_params,
            )
        else:
            raise AttributeError("FAIL TO LOAD DATALAODER")
        return dataloader

    def get_model(self) -> Optional[nn.Module]:
        """
        Return the pytorch model from config file.
        :return pytorch model from class in the config file.
        """
        model_class = HPKind.MODEL.get("class", None)
        params = HPKind.MODEL.get("params", {})
        if model_class is not None:
            return model_class(**params)
        else:
            raise AttributeError("MODEL NOT LOADED")

    def get_pl_model(self) -> Optional[AbstractPlModule]:
        """
        Return the pytorch lightning model from config file information.
        :return: the pytorch lightning model from the config file.
        """
        pl_model_class = HPKind.PLMODEL.get("class", None)
        if pl_model_class is not None:
            params = HPKind.PLMODEL.get("params", {})
            model = self.get_model()
            params["model"] = model
            checkpoint = params.get("checkpoint_path", None)
            # Load weight if checkpoint is in the params
            if checkpoint is not None and os.path.exists(checkpoint):
                pl_model = pl_model_class.load_from_checkpoint(**params)
                logger.debug(f"MODEL LOADED FROM {checkpoint}")
            else:
                pl_model = pl_model_class(**params)
            return pl_model
        else:
            raise AttributeError("PL MODEL NOT LOADED")

    def get_trainer(self) -> Optional[Trainer]:
        """Return the trainer form yaml file."""
        params = HPKind.TRAINER.get("params", {})
        if params is not None:
            trainer = Trainer(**params)
            return trainer
        else:
            logger.debug("Trainer not initialized")
            return None

    def get_log_dir(self) -> Optional[str]:
        """Return the directory where should be stored the logs."""
        log_dir = HPKind.LOG_DIR
        return log_dir

    def get_checkpoint_dir(self) -> Optional[str]:
        """Return where are stored the checkpoints."""
        chkpt_dir = HPKind.CHECKPOINT_DIR
        return chkpt_dir

    def get_model_name(self) -> str:
        """Return the name of the model if specified."""
        model_name = HPKind.MODEL_NAME
        model_name = "" if model_name is None else model_name
        return model_name

    def get_logger(self) -> List[LoggerAbstract]:
        """Return a list of logger."""
        loggers = HPKind.LOGGERS
        return loggers

    def save_checkpoints(self, log_path: str, checkpoint_path: str) -> None:
        """
        Save the hyperparameters file and add the model checkpoint.
        :param log_path: path to the logs
        :param checkpoint_path: path where is stored the model checkpoint.
        """
        path_to_save = os.path.join(log_path, "config.py")
        shutil.copyfile(inspect.getfile(HPKind), path_to_save)
        logger.debug(f"PYTHON CONFIG SAVED IN : {path_to_save}")



"""Dataset abstract file for Pytorch Dataset implementation."""
from abc import abstractmethod
from typing import List, Optional

from torch.utils.data import Dataset

from cthunder.features.preprocess_abstract import PreprocessAbstract
from cthunder.data.sample import Sample


class DatasetAbstract(Dataset):
    """Dataset abstract class that defines the main functions to use."""

    def __init__(
        self,
        preprocesses: Optional[List[PreprocessAbstract]] = None,
        device: Optional[str] = None,
        *args,
        **kwargs
    ):
        """ """
        self.preprocesses = preprocesses
        self.device = device
        self.setup_data(*args, **kwargs)

    @abstractmethod
    def read_data(self, item: int) -> Sample:
        """Return one element of the data (a line for a dataframe for instance)"""
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def setup_data(self, *args, **kwargs) -> None:
        """Initialise the data by loading the data (like dataframe, etc)"""
        raise NotImplementedError

    def preprocess(self, sample: Sample, *args, **kwargs) -> Sample:
        """
        Do the preprocessing.
        :param sample:
        :param args:
        :param kwargs:
        :return:
        """
        if self.preprocesses is not None:
            for process in self.preprocesses:
                sample = process.transform(sample, *args, **kwargs)
        return sample

    def __getitem__(self, item):
        """Should use at least 'self.preprocess(sample)'."""
        sample = self.read_data(item)
        sample = self.preprocess(sample)
        x, y = sample.read_to_torch(self.device)
        return x, y



import glob
import importlib
import json
import os
from typing import Any, Dict, List, Optional, Union, Tuple

import yaml
from loguru import logger


def open_yml(path: str) -> Optional[Dict]:
    """Open a .yml or .yaml file.

    :param path: the path where is stored the .yml/.yaml file.
    """
    if path.endswith(".yml") or path.endswith(".yaml"):
        with open(path) as file:
            try:
                result = yaml.safe_load(file)
                return result
            except yaml.YAMLError as exc:
                logger.debug(f"ERROR IN READING .YML FILE : {exc}")
                return None
    else:
        logger.debug(f"NOT A .YML or .YAML FILE : {path}")
        return None


def save_to_yaml(path: str, document: Dict) -> None:
    """
    Save a dictionary to yaml file.
    :param document: the json file to save
    :param path: path where to save the file.
    """
    if not (path.endswith(".yml") or path.endswith(".yaml")):
        path = path + ".yml"
    with open(path, "w") as yaml_file:
        yaml.dump(document, yaml_file)


def listdir_or_py(
    path: str, to_remove: List[str] = ["__pycache__", "__init__.py", "wandb"]
):
    """List the directory or python files.
    Args
        :param path: the path to get the directories from
    """
    # Get all not hidden python files.
    dirs = glob.glob(os.path.join(path, "**/*.py"), recursive=True)
    # Remove not wanted files or directories
    new_dirs = []
    for file in dirs:
        conditions = [file.endswith(x) for x in to_remove]
        if not any(conditions):
            new_dirs.append(file)
    return new_dirs


def instantiate_class_from_name(path: str, class_name: str) -> Optional[Any]:
    """Return the class from the name.

    :param path: the directory where to search the class
    :param class_name: the name of the class to instantiate
    """
    files = listdir_or_py(path)
    files_from_helper = listdir_or_py("thunder")
    files = files + files_from_helper
    for file in files:
        class_inst = get_class_from_file(file, class_name)
        if class_inst is not None:
            return class_inst
    return None


def get_class_from_file(python_file: str, class_name: str) -> Optional[Any]:
    """Return the instance class from the python file.

    :param python_file: the path to a python file
    :param class_name: the name of the class to instantiate
    :returns the class that is inside this file if it exists
    Example:
        if there is a class A in a file : example/folder/file.py
        the inputs are the following :
            python_file: example/folder/file.py
            class_name: "A"
    """
    module = python_file.replace(os.sep, ".").replace(".py", "")
    try:
        class_inst = getattr(importlib.import_module(module), class_name)
        return class_inst
    except (ImportError, AttributeError):
        return None


def read_json(path: str) -> Dict:
    """Read the json file.
    Args
    :path the path to the file
    :return a dictionary of the json file
    """
    if path.endswith(".json"):
        with open(path, "rb") as f:
            content = json.loads(f.read())
            return content
    else:
        logger.debug("Not a json file")
        return {}


def save_json(object: Dict, path: str):
    """Save the dictionary into a json file.
    Args
    :param object: the object to save
    :param path: the path where to save. Could have .json or not in the path
    """
    if path.endswith(".json"):
        path_to_save = path
    else:
        path_to_save = path + ".json"
    with open(path_to_save, "w") as file:
        json.dump(object, file)


def instantiate_class_from_init(init: Dict[str, Any]) -> Any:
    """Instantiates a class with the given args and init.
    Code original : "from lightning.pytorch.cli import instantiate_class"

    Args:
        init: Dict of the form {"class_path":...,"init_args":...}.

    Returns:
        The instantiated class object.
    """
    kwargs = init.get("init_args", {})
    class_module, class_name = init["class_path"].rsplit(".", 1)
    module = __import__(class_module, fromlist=[class_name])
    args_class = getattr(module, class_name)
    return args_class(**kwargs)



"""Class that does the main functions of a deep learning model."""
from typing import Dict, Union

import hydra
from loguru import logger
from omegaconf import DictConfig

from cthunder.config.config_helper_yaml import ConfigHelperYAML
from cthunder.experiments.abstract_experiment import AbstractExperiment


class CLIHelper:
    def __init__(self, config_path: Union[str, Dict], *args, **kwargs):
        self.config_path = config_path
        self.config_helper = ConfigHelperYAML(config_path)

    @staticmethod
    @hydra.main(version_base=None)
    def cli(cfg: DictConfig) -> None:
        """
        Launch the appropriate command.
        It will launch what is in the `command` parameter.
        Args:
            cfg: a dictionary with all the hyperparameters.
        """
        command = cfg.get("command", None)
        if command is None:
            logger.debug(
                "NO LAUNCHING COMMAND SPECIFIED. IT SHOULD BE IN THE 'command' name"
            )
        cli_helper = CLIHelper(config_path=cfg)
        experiment = cli_helper.config_helper.get_experiment_class()
        name_to_function = {
            "train": experiment.train,
            "test": cli_helper.test,
            "infer": cli_helper.infer,
            "predict": experiment.predict,
        }
        if command in name_to_function:
            name_to_function[command](**cfg)
        else:
            logger.debug(f"{command} IS NOT A COMMAND AVAILABLE")

    def train(self, experiment: AbstractExperiment, *args, **kwargs):
        """
        Do the training process
        :param experiment: the experiment to use
        :return:
        """
        experiment.train(*args, **kwargs)

    def test(self, experiment: AbstractExperiment, *args, **kwargs):
        """
        Do the evaluation process
        :param args:
        :param kwargs:
        :return:
        """
        experiment.test(*args, **kwargs)

    def infer(self, *args, **kwargs):
        """
        Do inference for a given input.
        :param args:
        :param kwargs:
        :return:
        """
        logger.debug("INFERENCE CLI")


if __name__ == "__main__":
    CLIHelper.cli()




"""Abstract Class for Experiment using Pytorch Lightning."""
import os
import random
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Type, Union

import numpy as np
import torch
from loguru import logger
from omegaconf import DictConfig
from lightning.pytorch import Callback, LightningDataModule, Trainer, LightningModule
from lightning.pytorch.callbacks import ModelCheckpoint, TQDMProgressBar
from lightning.pytorch.loggers import CSVLogger

from cthunder.config.config_helper_abstract import ConfigHelperAbstract
from cthunder.config.config_helper_python import ConfigHelperPython
from cthunder.config.config_helper_yaml import ConfigHelperYAML
from cthunder.data.datamodule_abstract import DataModuleAbstract
from cthunder.features.postprocess_abstract import PostprocessAbstract
from cthunder.features.preprocess_abstract import PreprocessAbstract
from cthunder.loggers.logger_abstract import LoggerAbstract
from cthunder.pl_model.abstract_pl_model import AbstractPlModule
from cthunder.utils.utils import save_json




@dataclass
class AbstractExperiment:
    """Abstract experiment to setup the training, evaluation or inference.
    variables:
        :param pl_model: the pytorch lightning model
        :param datamodule: the datamodule with the different dataloaders
        :param preprocesses: the different preprocesses to use
        :param postprocesses: the different postprocesses to use
        :param callbacks: the pytorch lightning callbacks
        :param trainer: the pytorch lightning trainer
        :param config: the config helper class to convert yaml hp

    """

    pl_model: Optional[Union[LightningModule, AbstractPlModule]] = None
    datamodule: Optional[Union[LightningDataModule, DataModuleAbstract]] = None
    preprocesses: Optional[List[PreprocessAbstract]] = None
    postprocesses: Optional[List[PostprocessAbstract]] = None
    callbacks: Optional[List[Callback]] = None
    trainer: Optional[Trainer] = None
    config: Optional[ConfigHelperAbstract] = None

    def __init__(self, *args, **kwargs):
        """
        Abstract experiment. Setup for the training, evaluation and inference.
        :param args:
        :param kwargs:
        """
        self.custom_loggers = None
        self.init_config_helper(*args, **kwargs)
        self.set_seed(seed=77)

    def set_seed(self, seed: int = 77):
        """
        Set the seed for numpy, random, torch
        """
        np.random.seed(seed)
        torch.manual_seed(seed)
        random.seed(seed)

    def init_config_helper(
            self,
            config_class: Type[ConfigHelperAbstract] = None,
            config_path: Optional[str] = None,
            *args,
            **kwargs,
    ):
        """Initialise the config helper to get the hyperparameters.
        Args:
            :param config_class: the class to instantiate the config helper
            :param config_path: path to the configuration file.
        """
        if config_path is not None:
            if isinstance(config_path, DictConfig):
                self.config_path = None
                self.config = ConfigHelperYAML(config_path)
                return None
            elif config_path.endswith(".py"):
                config_class = ConfigHelperPython
            elif config_path.endswith(".yml") or config_path.endswith(".yaml"):
                config_class = ConfigHelperYAML
            else:
                logger.debug(f"BAD EXTENSION FOR CONFIG FILE")
                return None
        self.config_path = config_path
        # Initialise the config helper with the path
        self.config = config_class(self.config_path)

    def train(self, *args, **kwargs):
        """Training loop."""
        self.setup_train(*args, **kwargs)
        self.trainer.fit(self.pl_model, self.datamodule)
        self.post_train(*args, **kwargs)

    def evaluate(self, *args, **kwargs):
        """Do the evaluation process."""
        self.setup_eval(*args, **kwargs)
        self.trainer.test(self.pl_model, self.datamodule)
        self.post_eval(*args, **kwargs)

    def validate(self, *args, **kwargs):
        """Do the validation process."""
        self.setup_eval(*args, **kwargs)
        self.trainer.validate(self.pl_model, self.datamodule)
        self.post_val(*args, **kwargs)

    def predict(self, path_to_save: str, *args, **kwargs):
        self.setup_eval(*args, **kwargs)
        if self.trainer is not None and self.datamodule is not None:
            predictions = self.trainer.predict(
                self.pl_model, self.datamodule.test_dataloader()
            )
            predictions = self._convert_preds_to_lists(predictions)
            save_json(predictions, path_to_save)  # type: ignore
        else:
            logger.debug("TRAINER IS NOT : COULD NOT PREDICT")

    def _convert_preds_to_lists(self, predictions: Any):
        """Convert the predictions from the batches to list."""
        y, preds = [x["y"] for x in predictions], [x["preds"] for x in predictions]
        y = np.squeeze(y).reshape(-1).tolist()
        preds = np.squeeze(preds).reshape(-1).tolist()
        output = {"y_true": y, "y_pred": preds}
        return output

    def post_val(self, *args, **kwargs):
        """Steps to do after validation."""
        pass

    def post_eval(self, *args, **kwargs):
        """Steps to do after evaluation (save predictions, ...)"""
        pass

    def post_train(self, *args, **kwargs):
        """Save the config file and change the checkpoint path."""
        log_path = os.path.join(
            self.trainer.logger.save_dir,
            "lightning_logs",
            f"version_{self.trainer.logger.version}",
        )
        self.config.save_checkpoints(log_path, self.checkpoints.best_model_path)

    def setup_eval(self, *args, **kwargs):
        """Do the setup for the evaluation process."""
        self.setup_train(*args, **kwargs)
        self.pl_model.eval()

    def setup_val(self, *args, **kwargs):
        """Do the setup for the validation."""
        self.setup_train(*args, **kwargs)

    def setup_train(self, *args, **kwargs):
        """Initialise the parameters for the training."""
        self.init_preprocesses(*args, **kwargs)
        self.init_postprocesses(*args, **kwargs)
        self.init_datamodule(*args, **kwargs)
        self.init_logger(*args, **kwargs)
        self.init_pl_model(*args, **kwargs)
        self.init_callbacks(*args, **kwargs)
        self.setup_trainer(*args, **kwargs)

    def init_pl_model(
            self,
            pl_model: Optional[AbstractPlModule] = None,
            custom_loggers: Optional[List[LoggerAbstract]] = None,
            *args,
            **kwargs,
    ):
        """Initialise the pytorch lightning model.
        :param pl_model: the pytorch lightning class that does the training process.
        """
        self.pl_model = self._init_module(pl_model, "pl_model", AbstractPlModule)
        if self.pl_model is not None:
            self.pl_model.config_path = self.config_path  # type: ignore

    def _init_module(
            self, module: Any, name: str, class_instance: Any = None, *args, **kwargs
    ):
        """
        Return the value if exists in the config file.
        :param module: the object to return if not None
        :param name: name of the element to return.
        :param class_instance: the instance of the class that module should be
        :return:
        """
        condition = (
            isinstance(module, class_instance)
            if class_instance is not None
            else (module is not None)
        )
        if condition:
            return module
        else:
            if self.config is not None:
                assert name in self.config.name_to_function
                return self.config.name_to_function.get(name)(*args, **kwargs)  # type: ignore
            else:
                logger.debug("METHOD NOT FOUND IN CONFIG HELPER")

    def init_datamodule(
            self, datamodule: Optional[DataModuleAbstract] = None, *args, **kwargs
    ):
        """
        Initialise the dataloaders.
        :param datamodule: the datamodule that returns the different dataloaders.
        """
        self.datamodule = self._init_module(
            datamodule, "datamodule", DataModuleAbstract
        )

    def init_preprocesses(
            self, preprocesses: Optional[List[PreprocessAbstract]] = None, *args, **kwargs
    ):
        """Initialise the preprocessing.
        Args:
            :param preprocesses: a list of different preprocesses.
        """
        self.preprocesses = self._init_module(preprocesses, "preprocesses")

    def init_postprocesses(
            self, postprocesses: Optional[List[PostprocessAbstract]] = None, *args, **kwargs
    ):
        """Initialise the postprocesses.
        Args:
             :param postprocesses: a list of different postprocesses.
        """
        self.postprocesses = self._init_module(postprocesses, "postprocesses")

    def init_callbacks(self, *args, **kwargs):
        """Initialise the callbacks from pytorch lightning module."""
        self.callbacks = []

    def init_logger(self, log_dir: Optional[str] = None, *args, **kwargs):
        """
        Initialise the logger.
        :param log_dir: directory where will be stored the logs.
        """
        log_dir = self._init_module(log_dir, "log_dir")
        log_dir = log_dir if log_dir is not None else "logs"
        logger = CSVLogger(save_dir=log_dir)
        return logger

    def setup_trainer(self, trainer_params: Optional[Dict] = None, *args, **kwargs):
        """
        Set the trainer (from pytorch lightning module)
        :param trainer_params: parameters for the trainer.
        """
        self.checkpoints = self.init_model_checkpoint_callback(*args, **kwargs)
        progress_bar = self.init_progress_bar(*args, **kwargs)
        callbacks = [*self.callbacks, *[self.checkpoints, progress_bar]]  # type: ignore
        logger = self.init_logger(*args, **kwargs)
        if trainer_params is not None:
            # Initialise with current parameters
            trainer = Trainer(**trainer_params)
        elif trainer_params is None and self.config is not None:
            # Initialise with the config helper
            trainer = self.config.get_trainer()  # type: ignore
        else:
            raise AttributeError("TRAINER CAN'T BE INSTANTIATE")
        trainer.callbacks = callbacks
        trainer.logger = logger
        self.trainer = trainer

    def init_progress_bar(self, *args, **kwargs):
        """Init the progress bar."""
        return TQDMProgressBar()

    def init_model_checkpoint_callback(
            self,
            checkpoint_dir: Optional[str] = "models",
            monitor_loss: Optional[str] = "val_loss",
            model_name: Optional[str] = "example",
            save_top_k: Optional[int] = 1,
            *args,
            **kwargs,
    ):
        """Initialise the callback that specify where are stored the checkpoints."""
        if self.config is not None:
            checkpoint_dir = self._init_module(checkpoint_dir, "checkpoint_dir")
            model_name = self._init_module(model_name, "model_name")
            monitor_loss = "val_loss" if monitor_loss is None else monitor_loss
            save_top_k = 1 if save_top_k is None else save_top_k
        filename = f"{model_name}-{{{monitor_loss}:.2f}}-{{epoch}}"
        return ModelCheckpoint(
            dirpath=checkpoint_dir,
            filename=filename,
            monitor=monitor_loss,
            mode="min" if "loss" in monitor_loss else "max",  # type: ignore
            save_top_k=save_top_k,  # type: ignore
        )

    def setup_inference(self, *args, **kwargs):
        """Setup the model for inference."""
        pass

    def infer(self, *args, **kwargs):
        """Do the inference."""
        pass



"""Class that convert a file with hyperparameters to python readable variables."""
from abc import abstractmethod
from typing import Any, List, Optional, Union

from omegaconf import DictConfig
from pytorch_lightning import Trainer
from torch import nn

from cthunder.data.datamodule_abstract import DataModuleAbstract
from cthunder.data.dataset_abstract import DatasetAbstract
from cthunder.features.postprocess_abstract import PostprocessAbstract
from cthunder.features.preprocess_abstract import PreprocessAbstract
from cthunder.loggers.logger_abstract import LoggerAbstract
from cthunder.pl_model.abstract_pl_model import AbstractPlModule



class ConfigHelperAbstract:
    def __init__(self, config_path: Union[str, DictConfig], *args, **kwargs):
        """Initialise the helper with a file.

        Args:
            :param config_path: the path to a configuration file.
        """
        self.config_path = config_path
        self.config = self.read_config(self.config_path)
        # Dictionary to convert names to given functions.
        self.name_to_function = {
            "preprocesses": self.get_preprocesses,
            "postprocesses": self.get_postprocesses,
            "dataset": self.get_dataset,
            "datamodule": self.get_datamodule,
            "model": self.get_model,
            "pl_model": self.get_pl_model,
            "trainer": self.get_trainer,
            "log_dir": self.get_log_dir,
            "checkpoint_dir": self.get_checkpoint_dir,
            "model_name": self.get_model_name,
            "logger": self.get_logger,
        }

    @abstractmethod
    def read_config(self, config_path: Union[str, DictConfig]) -> Any:
        """Read the content of the config file."""
        raise NotImplementedError

    @abstractmethod
    def get_preprocesses(self) -> Optional[List[PreprocessAbstract]]:
        """
        Return the preprocesses from the config file.
        :return: a list of preprocesses initialised with they parameters
        """
        raise NotImplementedError

    @abstractmethod
    def get_postprocesses(self) -> Optional[List[PostprocessAbstract]]:
        """
        Return the postprocesses from the config file.
        :return: a list of postprocesses initialised with they parameters
        """
        raise NotImplementedError

    @abstractmethod
    def get_dataset(self) -> Optional[DatasetAbstract]:
        """
        Return the Dataset class from the config file.
        :return: the Dataset instantiate.
        """
        raise NotImplementedError

    @abstractmethod
    def get_datamodule(self) -> DataModuleAbstract:
        """
        Return the DataLoader class from the config file.
        It initialises the datasets and preprocesses
        :return: the DataLoader instantiate.
        """
        raise NotImplementedError

    @abstractmethod
    def get_model(self) -> Optional[nn.Module]:
        """
        Return the pytorch model from config file.
        :return pytorch model from class in the config file.
        """
        raise NotImplementedError

    @abstractmethod
    def get_pl_model(self) -> Optional[AbstractPlModule]:
        """
        Return the pytorch lightning model from config file.
        :return: the pytorch lightning model from the config file.
        """
        raise NotImplementedError

    @abstractmethod
    def get_trainer(self) -> Optional[Trainer]:
        """Return the trainer from config file."""
        raise NotImplementedError

    @abstractmethod
    def get_log_dir(self) -> Optional[str]:
        """Return the directory where should be stored the logs."""
        raise NotImplementedError

    @abstractmethod
    def get_checkpoint_dir(self) -> Optional[str]:
        """Return where are stored the checkpoints."""
        raise NotImplementedError

    @abstractmethod
    def get_model_name(self) -> str:
        """Return the name of the model if specified."""
        raise NotImplementedError

    @abstractmethod
    def get_logger(self) -> List[LoggerAbstract]:
        """Return a list of logger."""
        raise NotImplementedError

    @abstractmethod
    def save_checkpoints(self, log_path: str, checkpoint_path: str) -> None:
        """
        Save the hyperparameters file and add the model checkpoint.
        :param log_path: path to the logs
        :param checkpoint_path: path where is stored the model checkpoint.
        """
        raise NotImplementedError

    @staticmethod
    def get_experiment_class(self) -> Any:
        """
        Return the class given by the name in the config file
        """
        raise NotImplementedError



"""Class that implements Dataloader Abstract"""
from abc import abstractmethod
from typing import Dict, List, Optional, Union

import torch
from lightning.pytorch import LightningDataModule
from torch.utils.data import DataLoader, Dataset

from cthunder.features.preprocess_abstract import PreprocessAbstract
from cthunder.utils.utils import instantiate_class_from_init


class DataModuleAbstract(LightningDataModule):
    """
    Class that loads the different dataloaders for training, validation and testing sets.

    Methods:
        __init__(self, preprocesses, batch_size, shuffle, device, *args, **kwargs)
        prepare_data(self)
        collate_fn(self)
        _get_dataloader_from_split(self, split)
        train_dataloader()
        val_dataloader()
        test_dataloader()
    """

    def __init__(
        self,
        dataset_init: Dict,
        preprocesses: Optional[List[PreprocessAbstract]],
        batch_size: int,
        device: Union[Optional[str], torch.device],
        shuffle: bool = False,
        num_workers: int = 0,
        *args,
        **kwargs,
    ):
        """
        Initialisation of the class
        Args:
        :param dataset_init: a dictionary with the "class_path" and "init_args" to instantiate
            the dataset
        :param preprocesses: the list of different preprocessing to do
        :param batch_size: the number of examples for a batch
        :param shuffle: whether to shuffle or not the train dataset.
        :param device: which device to use
        :param num_workers: the number of workers for the DataLoader.

        Attributes:
        all_shuffle : dictionary of shuffle value for training, validation and test sets.
        datasets: dictionary with Dataset for training, validation and test sets.
        """
        super().__init__()
        self.dataset_init = dataset_init
        self.preprocesses = preprocesses
        self.batch_size = batch_size
        self.device = device
        self.num_workers = num_workers
        self.datasets: Dict[str, Dataset] = {
            "train": Dataset(),
            "valid": Dataset(),
            "test": Dataset(),
        }
        self.all_shuffle = {"train": True, "valid": False, "test": False}
        self.prepare_data(*args, **kwargs)

    @staticmethod
    def instantiate_dataset(dataset_init: Dict, *args, **kwargs) -> Dataset:
        """Instantiate the dataset with the given arguments.

        Args:
             :param dataset_init: Dict of the form {'class_path': ..., 'init_args': ...}
        :return an instance of the Dataset
        """
        dataset_init["init_args"] = {**dataset_init.get("init_args", {}), **kwargs}
        return instantiate_class_from_init(init=dataset_init)

    @abstractmethod
    def prepare_data(self, *args, **kwargs):
        """Init the train, valid and test datasets.
            Should fill the dictionary self.datasets
            Should download the data if necessary.

        Example :
            self.datasets['train'] = DatasetAbstract(*args, **kwargs)
            self.datasets['valid'] = DatasetAbstract(*args, **kwargs)
            self.datasets['test'] = DatasetAbstract(*args, **kwargs)

        """
        raise NotImplementedError

    @staticmethod
    def collate_fn(batch, *args, **kwargs):
        """Collate function to merge element to create a batch."""
        return None

    def _get_dataloader_from_split(self, split: str) -> DataLoader:
        """
        Returns the dataloader from the self.datasets with the needed arguments.

        :param split: which dataset (train, valid or test)
        :return a dataloader with the hyperparameters from the initialisation
        :raise NotImplementedError if the self.datasets has not been implemented
        """
        # Check the class of 'collate_fn' : if this is from abstract class, return None
        if self.collate_fn.__qualname__.split(".")[0] == "DataModuleAbstract":
            collate_fn = None
        else:
            collate_fn = self.collate_fn
        if split in self.datasets:
            return DataLoader(
                self.datasets[split],  # type : ignore
                batch_size=self.batch_size,
                shuffle=self.all_shuffle.get(split, False),
                collate_fn=collate_fn,
                num_workers=self.num_workers,
                persistent_workers=True
            )
        else:
            raise NotImplementedError(f"Split not in dataset keys : {split}")

    def train_dataloader(self):
        """Return the training dataloader."""
        return self._get_dataloader_from_split("train")

    def val_dataloader(self):
        """Return the validation dataloader."""
        return self._get_dataloader_from_split("valid")

    def test_dataloader(self):
        """Return the testing dataloader."""
        return self._get_dataloader_from_split("test")



"""Preprocess abstract class to define the main function to use."""
from abc import abstractmethod

from cthunder.data.sample import Sample


class PreprocessAbstract:
    """Class that represents the different preprocessing."""

    @abstractmethod
    def transform(self, obj: Sample, *args, **kwargs):
        """Preprocess the data."""
        raise NotImplementedError



"""File for the sample, which describes the content of data."""
from abc import abstractmethod
from typing import Dict, List, Optional, Union

import numpy as np
import torch


class Sample:
    """
    Class that represents data.
    """

    def read_to_torch(self, device: Optional[str], *args, **kwargs):
        """Read the data to torch format."""
        x, y = self.convert_to_x_and_y()
        x, y = self.convert_to_torch(x, device), self.convert_to_torch(y, device)
        return x, y

    @abstractmethod
    def convert_to_x_and_y(self):
        """Convert the sample into X and y"""
        raise NotImplementedError

    @staticmethod
    def _get_device(device: Optional[str]) -> torch.device:
        """
        Return torch.device of the given device
        :param device: the device for pytorch code.
        :return: torch device
        """
        if isinstance(device, str):
            torch_device = torch.device(device)
        elif isinstance(device, torch.device):
            torch_device = device
        else:
            torch_device = torch.device("cpu")
        return torch_device

    @staticmethod
    def convert_to_torch(
        inputs: Union[np.ndarray, List, torch.Tensor, Dict], device: Optional[str]
    ) -> Union[torch.Tensor, Dict]:
        """Convert the inputs into tensor

        :param inputs: elements to be converted to torch format.
        :param device: the device for pytorch code.
        """
        torch_device = Sample._get_device(device)
        if isinstance(inputs, list):
            output = torch.tensor(inputs).to(torch_device)
        elif isinstance(inputs, np.ndarray):
            output = torch.from_numpy(inputs).to(torch_device)
        elif isinstance(inputs, torch.Tensor):
            output = inputs.to(torch_device)
        elif isinstance(inputs, Dict):
            output = Sample._convert_to_torch_dict(inputs, device)  # type: ignore
        else:
            raise NotImplementedError(
                f"Conversion to torch not possible for type : {type(inputs)}"
            )
        return output

    @staticmethod
    def _convert_to_torch_dict(inputs: Dict, device: str) -> Dict:
        """
        Convert the dict inputs to torch.
        :param inputs: dictionary of elements to be converted to torch.
        :return: dictionary with torch elements as values
        """
        new_inputs = {}
        for key, value in inputs.items():
            if isinstance(value, Dict):
                raise TypeError("Nested Dictionary found for conversion")
            else:
                new_inputs[key] = Sample.convert_to_torch(value, device)
        return new_inputs



"""Class that maps the information from config.yaml file."""
import os.path
from typing import Any, Dict, List, Optional, Union

from loguru import logger
from omegaconf import DictConfig, OmegaConf
from lightning import Trainer
from torch import nn

from cthunder.config.config_helper_abstract import ConfigHelperAbstract
from cthunder.data.datamodule_abstract import DataModuleAbstract
from cthunder.data.dataset_abstract import DatasetAbstract

from cthunder.features.preprocess_abstract import PreprocessAbstract
from cthunder.loggers.logger_abstract import LoggerAbstract
from cthunder.pl_model.abstract_pl_model import AbstractPlModule
from cthunder.utils.utils import open_yml, save_to_yaml, instantiate_class_from_init



class ConfigHelperYAML(ConfigHelperAbstract):
    """Class that return the main elements from the yaml config file."""

    def __init__(self, config_path: Union[str, DictConfig], *args, **kwargs):
        """
        Load the config.yml or config.yaml file.
        :param path_to_config: the path to the yaml config file or a dict with the given parameters
        """
        super().__init__(config_path, *args, **kwargs)
        self.init_path()

    def init_path(self):
        """Initialise the path where to search the instances of class."""
        self.path = self.config.get("path", "")

    def read_config(self, config_path: Union[str, DictConfig]):
        """Read the content of the config file."""
        if isinstance(config_path, str):
            config_file = open_yml(config_path)
        elif isinstance(config_path, DictConfig):
            config_file = config_path
        else:
            logger.debug("NO CONFIGURATION PATH")
            return None
        if config_file is not None:
            return config_file
        else:
            logger.debug("CONFIG PATH NOT READABLE")

    def _get_value_from_keys(
            self, keys: Union[List[str], str], config: Optional[Dict] = None
    ) -> Any:
        """
        Return the value from nested keys.
        :param keys: List of keys in the config file
        :param config: the dictionary where to search the elements.
        :return: the value corresponding.
        """
        config = config if config is not None else self.config
        if isinstance(keys, str):
            output = config.get(keys, None)
            return dict(output) if output is not None else output
        elif isinstance(keys, list):
            if len(keys) == 1:
                output = config.get(keys[0], None)
                return dict(output) if output is not None else output
            else:
                output = self._get_value_from_keys(
                    keys[1:],
                    config.get(keys[0], {}),
                )
                return dict(output) if output is not None else output
        else:
            raise AttributeError(
                "KEYS TO SEARCH IN CONFIG FILE BAD TYPE : SHOULD BE STR OR LIST, "
                f"AND GOT VALUE OF TYPE {type(keys)} : {keys}"
            )

    def _get_processes(self, name_process: str) -> Optional[List]:
        """
        Return the preprocesses or postprocesses from the yaml config file.
        :param name_process: either postprocesses or preprocesses
        :return: a list of processes with their parameters
        """
        processes: List = []
        list_processes = self._get_value_from_keys([name_process])
        if list_processes is None:
            return processes
        for name in list_processes:
            params = self._get_value_from_keys([name_process, name])
            class_inst = instantiate_class_from_init(init=params)
            if class_inst is not None:
                processes.append(class_inst)
        return processes

    def get_preprocesses(self) -> Optional[List[PreprocessAbstract]]:
        """
        Return the preprocesses from the yaml config file.
        :return: a list of preprocesses initialised with their parameters
        """
        return self._get_processes("preprocesses")

    def get_postprocesses(self) -> Optional[List[PreprocessAbstract]]:
        """
        Return the preprocesses from the yaml config file.
        :return: a list of preprocesses initialised with their parameters
        """
        return self._get_processes("postprocesses")

    def get_dataset(self) -> Optional[DatasetAbstract]:
        """
        Return the Dataset class from the yaml config file.
        This function doesn't use the hyperparameters of the dataset.
        :return: the Dataset instantiate.
        """
        dataset_params = self._get_value_from_keys(["dataset"])
        if dataset_params is None:
            return None
        dataset_class = instantiate_class_from_init(init=dataset_params)
        if dataset_class is not None:
            return dataset_class
        else:
            raise AttributeError("FAIL TO LOAD DATASET")

    def get_datamodule(self) -> DataModuleAbstract:
        """
        Return the DataLoader class from the yaml config file.
        It initialises the datasets and preprocesses
        It gets also the hyperparameters of the dataset.
        :return: the DataLoader instantiate.
        """
        # Get the preprocesses
        preprocess = self.get_preprocesses()
        dataset_params = self._get_value_from_keys(["dataset"])
        # Get the dataloader params and class
        datamodule_params = self._get_value_from_keys(["datamodule"])
        # Fill the last parameters : preprocesses and dataset class
        params = {"preprocesses": preprocess, "dataset_init": dataset_params}
        datamodule_params["init_args"] = {
            **datamodule_params.get("init_args", {}),
            **params,
        }
        datamodule = instantiate_class_from_init(init=datamodule_params)
        return datamodule

    def get_model(self) -> Optional[nn.Module]:
        """
        Return the pytorch model from yaml.
        :return pytorch model from class in the config file.
        """
        model_params = self._get_value_from_keys("model")
        if model_params is None:
            return None
        model = instantiate_class_from_init(model_params)
        return model

    def get_pl_model(self) -> Optional[AbstractPlModule]:
        """
        Return the pytorch lightning model from yaml information.
        :return: the pytorch lightning model from the config file.
        """
        pl_model_params = self._get_value_from_keys(["pl_model"])
        if pl_model_params is None:
            return None
        model = self.get_model()
        postprocess = self.get_postprocesses()
        loggers = self.get_logger()
        pl_model_params["init_args"] = {
            **pl_model_params.get("init_args", {}),
            **{"model": model},
            **{"postprocesses": postprocess},
            **{"custom_loggers": loggers},
        }
        chkpt_path = self.config.get("pl_model",{}).get("init_args", {}).get("checkpoint_path", None)
        if chkpt_path is None:
            pl_model = instantiate_class_from_init(pl_model_params)
        else:
            pl_model = instantiate_class_from_init(pl_model_params)
            pl_model = pl_model.load_from_checkpoint(**pl_model_params["init_args"])
        return pl_model

    def get_trainer(self) -> Trainer:
        """Return the trainer form yaml file."""
        params = self._get_value_from_keys(["trainer"])
        if params is not None:
            trainer = instantiate_class_from_init(params)
        else:
            trainer = Trainer()
        return trainer

    def get_log_dir(self) -> Optional[str]:
        """Return the directory where should be stored the logs."""
        log_dir = self._get_value_from_keys(["log_dir"])
        return log_dir

    def get_checkpoint_dir(self) -> Optional[str]:
        """Return where are stored the checkpoints."""
        chkpt_dir = self._get_value_from_keys(["checkpoint_dir"])
        return chkpt_dir

    def get_model_name(self) -> str:
        """Return the name of the model if specified."""
        model_name = self._get_value_from_keys(["model_name"])
        model_name = "" if model_name is None else model_name
        return model_name

    def get_logger(self) -> List[LoggerAbstract]:
        """Return a list of logger."""
        params = self.config.get("loggers", {})
        loggers = []
        for name, param in params.items():
            param_ = dict(param)
            init_args = dict(param_.get("init_args", {}))
            init_args['config'] = self.config
            param_['init_args'] = init_args
            loggers.append(instantiate_class_from_init(init=param_))
        return loggers

    def save_checkpoints(self, log_path: str, checkpoint_path: str) -> None:
        """
        Save the hyperparameters file and add the model checkpoint.
        :param log_path: path to the logs
        :param checkpoint_path: path where is stored the model checkpoint.
        """
        params = dict(self.config.get("pl_model", {}).get("init_args", {}))
        params["checkpoint_path"] = checkpoint_path
        path_to_save = os.path.join(log_path, "config.yml")
        self.config.get("pl_model", {})['init_args'] = params
        config = OmegaConf.to_container(self.config)
        save_to_yaml(path_to_save, config)
        logger.debug(f"YAML CONFIG SAVED IN : {path_to_save}")

    def get_experiment_class(self, *args, **kwargs) -> Any:
        """
        Return the class given by the name in the config file
        """
        params = self.config.get("experiment", None)
        if params is None:
            raise NotImplementedError("NO EXPERIMENT CLASS FOUND IN CONFIG")
        params = dict(params)
        params["init_args"] = {
            **params.get("init_args", {}),
            **{"config_path": self.config_path},
        }
        experiment = instantiate_class_from_init(init=params)
        return experiment
